/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */


#ifndef PEP_C_INTERACT
 #define PEP_C_INTERACT


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

/* include lcmaps_c_pep_config.h for PEP_API */
#include "lcmaps_c_pep_config.h"

#include <lcmaps/lcmaps_account.h>
#include <lcmaps/lcmaps_arguments.h>
#include <lcmaps/lcmaps_cred_data.h>

#if PEP_API == 1
#include <pep/pep.h>
#else
#include <argus/pep.h>
#endif

#include "xacml_attr_obl_profile.h"


#define  C_PEP_OK      1
#define  C_PEP_TRUE    1
#define  C_PEP_FAIL    2
#define  C_PEP_FALSE   2


#ifndef XACML_GRIDWN_PROFILE_VERSION
    #define XACML_GRIDWN_PROFILE_VERSION "http://glite.org/xacml/profile/grid-wn/1.0"
#endif
#ifndef XACML_AUTHZ_INTEROP_PROFILE_VERSION
    #define XACML_INTEROP_PROFILE_VERSION "http://authz-interop.org/profile/1.1"
#endif

#define PROFILE_AUTHZ_INTEROP 1
#define PROFILE_SWITCH_GRIDWN 2


void setPEPdurl (char *);
char * getPEPdURL (void);

#if PEP_API == 1
int pepc_initialize(int, pep_obligationhandler_t **, size_t *);

int pepc_construct_request (int use_profile,
                            const char * subject_proxy,
                            const char * actiontype,
                            const char * actionid,
                            const char * resourceid,
                            const char * resourcetype,
                            const char * dnshostname,
                            const char * environ_proxy);

int pepc_engage(void);

int pepc_cleanup (void);
#else
PEP *pepc_initialize(int, pep_obligationhandler_t **, size_t *); 

int pepc_construct_request (PEP *pep_handle,
			    int use_profile,
                            const char * subject_proxy,
                            const char * actiontype,
                            const char * actionid,
                            const char * resourceid,
                            const char * resourcetype,
                            const char * dnshostname,
                            const char * environ_proxy);

int pepc_engage(PEP *pep_handle);

int pepc_cleanup (PEP *pep_handle);
#endif


int pepc_construct_request_subject    (xacml_request_t ** request,
                                       const char * subject_proxy);
int pepc_construct_request_action     (xacml_request_t ** request,
                                       const char * actiontype,
                                       const char * actionid);
int pepc_construct_request_resource   (xacml_request_t ** request,
                                       const char * resourceid,
                                       const char * resourcetype,
                                       const char * dnshostname);
int pepc_construct_request_environment(xacml_request_t ** request,
                                       const char * environ_proxy);
int pepc_get_output (lcmaps_account_info_t **);


#endif /* PEP_C_INTERACT */
