/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include <sys/types.h>
#include <pwd.h>
#include <grp.h>

#include <lcmaps/lcmaps_log.h>

/* include lcmaps_c_pep_config.h for PEP_API */
#include "lcmaps_c_pep_config.h"

#if PEP_API == 1
#include <pep/pep.h>
#else
#include <argus/pep.h>
#endif

#include "pep-c-interact.h"



#ifndef PEP_C_OBLIGATION_HANDLERS_HELPERS_H
    #define PEP_C_OBLIGATION_HANDLERS_HELPERS_H

typedef struct list_supported_obligations_s {
    char *                                obligation_id;
    struct list_supported_obligations_s * next;
} list_supported_obligations_t;

/* setter and getter functions for treat_notapplicable_as_success option */
void set_treat_notapplicable_as_success(int val);
int get_treat_notapplicable_as_success(void);

char * get_next_supported_obligation_id (void);
void rewind_supported_obligation_list (void);
int add_supported_obligation_to_list (const char * obligation_id);
int free_supported_obligations_list (void);
int print_supported_obligations_list (void);

lcmaps_account_info_t * getAccountInfo (void);
const char * decision_str(xacml_decision_t);
const char * fulfillon_str(xacml_fulfillon_t);
int addCredentialDataFromUsername (const char * username);
int checkResponseSanity (xacml_response_t * response);

#if 0
int get_gidlist(const char * username, int * ngroups, gid_t ** group_list);
#endif
int threadsafe_getuid_from_name (const char * username, uid_t *uid);
int threadsafe_getgid_from_name (const char * groupname, gid_t *gid);


#endif /* PEP_C_OBLIGATION_HANDLERS_HELPERS_H */
